/* scherzo - Music notation training
 *
 * 	pitch.c - Common structures and functions for pitches, etc.
 *
 * Copyright © 2013 Carl Worth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/ .
 */

#include <stdlib.h>

#include "pitch.h"

const char *
pitch_string (pitch_t pitch)
{
    static char double_flat[] = "X𝄫";
    static char flat[] = "X♭";
    static char natural[] = "X";
    static char sharp[] = "X♯";
    static char double_sharp[] = "X𝄪";
    char *ret;

    switch (PITCH_ACCIDENTAL (pitch)) {
    case PITCH_ACCIDENTAL_DOUBLE_FLAT:
	ret = double_flat;
	break;
    case PITCH_ACCIDENTAL_FLAT:
	ret = flat;
	break;
    case PITCH_ACCIDENTAL_NATURAL:
	ret = natural;
	break;
    case PITCH_ACCIDENTAL_SHARP:
	ret = sharp;
	break;
    case PITCH_ACCIDENTAL_DOUBLE_SHARP:
	ret = double_sharp;
	break;
    }

    switch (PITCH_NAME (pitch)) {
    case PITCH_NAME_C:
	ret[0] = 'C';
	break;
    case PITCH_NAME_D:
	ret[0] = 'D';
	break;
    case PITCH_NAME_E:
	ret[0] = 'E';
	break;
    case PITCH_NAME_F:
	ret[0] = 'F';
	break;
    case PITCH_NAME_G:
	ret[0] = 'G';
	break;
    case PITCH_NAME_A:
	ret[0] = 'A';
	break;
    case PITCH_NAME_B:
	ret[0] = 'B';
	break;
    }

    return ret;
}

pitch_t
pitch_raise_by_octaves (pitch_t pitch, int octaves)
{
    int new_octave = PITCH_OCTAVE (pitch) + octaves;

    if (new_octave > 8)
	new_octave = 8;

    return PITCH (PITCH_NAME (pitch), PITCH_ACCIDENTAL (pitch), new_octave);
}

pitch_t
pitch_lower_by_octaves (pitch_t pitch, int octaves)
{
    int new_octave = PITCH_OCTAVE (pitch) - octaves;

    if (new_octave < 0)
	new_octave = 0;

    return PITCH (PITCH_NAME (pitch), PITCH_ACCIDENTAL (pitch), new_octave);
}

/* Number of half steps from 'root' up to 'pitch' (that is, counting
 * the steps up a chromatic scale), within the same octave. */
int
pitch_from_root_in_half_steps (pitch_t pitch, pitch_t root)
{
    int root_midi = pitch_to_midi (root);
    int pitch_midi = pitch_to_midi (pitch);

    while (pitch_midi < root_midi)
	pitch_midi += 12;

    return (pitch_midi - root_midi) % 12;
}

pitch_t
pitch_spell_as_degree (pitch_t pitch, pitch_t root, int degree)
{
    pitch_name_t name = PITCH_NAME (pitch);
    pitch_accidental_t accidental = PITCH_ACCIDENTAL (pitch);
    int octave = PITCH_OCTAVE (pitch);
    int degree_zero_based = (degree - 1) % 7;
    int degree_delta;

    int note_degree_zero_based = name - PITCH_NAME (root);
    if (note_degree_zero_based < 0)
	note_degree_zero_based += 7;

    if (note_degree_zero_based == degree_zero_based)
	return pitch;

    degree_delta = note_degree_zero_based - degree_zero_based;
    if (degree_delta > 3)
	degree_delta = - (7 - degree_delta);
    if (degree_delta < -3)
	degree_delta = - (-7 - degree_delta);

    /* Cannot re-spell pitch more than one degree away. Return
     * original pitch. */
    if (abs (degree_delta) != 1)
	return pitch;

    if (degree_delta == -1) {
	if (name == PITCH_NAME_B) {
	    name = PITCH_NAME_C;
	    octave++;
	} else {
	    name++;
	}
	switch (name) {
	case PITCH_NAME_D:
	case PITCH_NAME_E:
	case PITCH_NAME_G:
	case PITCH_NAME_A:
	case PITCH_NAME_B:
	    accidental -= 2;
	    break;
	case PITCH_NAME_C:
	case PITCH_NAME_F:
	    accidental -= 1;
	    break;
	}
    }

    if (degree_delta == +1) {
	if (name == PITCH_NAME_C) {
	    name = PITCH_NAME_B;
	    octave--;
	} else {
	    name--;
	}
	switch (name) {
	case PITCH_NAME_C:
	case PITCH_NAME_D:
	case PITCH_NAME_F:
	case PITCH_NAME_G:
	case PITCH_NAME_A:
	    accidental += 2;
	    break;
	case PITCH_NAME_E:
	case PITCH_NAME_B:
	    accidental += 1;
	}
    }

    /* Also cannot use accidentals to respell more than two half steps
     * either direction. Return original pitch. */
    if (accidental < 0 || accidental > PITCH_ACCIDENTAL_DOUBLE_SHARP)
	return pitch;

    return PITCH (name, accidental, octave);
}

/* Return a MIDI note value corresponding to 'pitch' */
unsigned char
pitch_to_midi (pitch_t pitch)
{
    int octave = PITCH_OCTAVE (pitch);
    unsigned char midi_note = 12 * (octave + 1);

    switch (PITCH_NAME (pitch)) {
    case PITCH_NAME_C:
	break;
    case PITCH_NAME_D:
	midi_note += 2;
	break;
    case PITCH_NAME_E:
	midi_note += 4;
	break;
    case PITCH_NAME_F:
	midi_note += 5;
	break;
    case PITCH_NAME_G:
	midi_note += 7;
	break;
    case PITCH_NAME_A:
	midi_note += 9;
	break;
    case PITCH_NAME_B:
	midi_note += 11;
	break;
    }

    switch (PITCH_ACCIDENTAL (pitch)) {
    case PITCH_ACCIDENTAL_DOUBLE_FLAT:
	midi_note -= 2;
	break;
    case PITCH_ACCIDENTAL_FLAT:
	midi_note -= 1;
	break;
    case PITCH_ACCIDENTAL_NATURAL:
	break;
    case PITCH_ACCIDENTAL_SHARP:
	midi_note += 1;
	break;
    case PITCH_ACCIDENTAL_DOUBLE_SHARP:
	midi_note += 2;
	break;
    }

    return midi_note;
}

/* Return the pitch_t value corresponding to the given MIDI note value. */
pitch_t
pitch_from_midi (unsigned char midi_note)
{
    int octave = octave = midi_note / 12 - 1;

    switch (midi_note % 12)
    {
    default:
    case 0:
	return PITCH_LITERAL (C, NATURAL, octave);
	break;
    case 1:
	return PITCH_LITERAL (C, SHARP, octave);
	break;
    case 2:
	return PITCH_LITERAL (D, NATURAL, octave);
	break;
    case 3:
	return PITCH_LITERAL (D, SHARP, octave);
	break;
    case 4:
	return PITCH_LITERAL (E, NATURAL, octave);
	break;
    case 5:
	return PITCH_LITERAL (F, NATURAL, octave);
	break;
    case 6:
	return PITCH_LITERAL (F, SHARP, octave);
	break;
    case 7:
	return PITCH_LITERAL (G, NATURAL, octave);
	break;
    case 8:
	return PITCH_LITERAL (G, SHARP, octave);
	break;
    case 9:
	return PITCH_LITERAL (A, NATURAL, octave);
	break;
    case 10:
	return PITCH_LITERAL (A, SHARP, octave);
	break;
    case 11:
	return PITCH_LITERAL (B, NATURAL, octave);
	break;
    }
}
