/* scherzo - Music notation training
 *
 * 	pitch.h - Common structures and functions for pitches, etc.
 *
 * Copyright © 2010,2013 Carl Worth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/ .
 */

#include <stdint.h>

#ifndef PITCH_H
#define PITCH_H

typedef uint32_t pitch_t;

#define PITCH_ACCIDENTAL_SHIFT	(0)
#define PITCH_ACCIDENTAL_MASK	(0x07 << PITCH_ACCIDENTAL_SHIFT)

#define PITCH_NAME_SHIFT	(3)
#define PITCH_NAME_MASK		(0x07 << PITCH_NAME_SHIFT)

#define PITCH_OCTAVE_SHIFT	(6)
#define PITCH_OCTAVE_MASK	(0x07 << PITCH_OCTAVE_SHIFT)

#define PITCH_ACCIDENTAL(pitch)	\
    (((pitch) & PITCH_ACCIDENTAL_MASK) >> PITCH_ACCIDENTAL_SHIFT)
#define PITCH_NAME(pitch)	\
    (((pitch) & PITCH_NAME_MASK) >> PITCH_NAME_SHIFT)
#define PITCH_OCTAVE(pitch)	\
    (((pitch) & PITCH_OCTAVE_MASK) >> PITCH_OCTAVE_SHIFT)

#define PITCH(name, accidental, octave)		  \
    (((octave) << PITCH_OCTAVE_SHIFT)		| \
     ((name) << PITCH_NAME_SHIFT)		| \
     ((accidental) << PITCH_ACCIDENTAL_SHIFT))

#define PITCH_LITERAL(literal_name, literal_accidental, octave)	\
    PITCH(PITCH_NAME_##literal_name,				\
	  PITCH_ACCIDENTAL_##literal_accidental,		\
	  octave)

#define PITCH_NATURAL(literal_name, octave) \
    PITCH_LITERAL(literal_name, NATURAL, octave)

/* PITCH_CLASS is useful for comparing pitches while ignoring any octave. */
#define PITCH_CLASS(name, accidental) PITCH(name, accidental, 0)

#define PITCH_CLASS_LITERAL(literal_name, literal_accidental) \
    PITCH_LITERAL(literal_name, literal_accidental, 0)

#define PITCH_NOT_A_PITCH ((pitch_t) -1)

typedef enum pitch_accidental
{
    PITCH_ACCIDENTAL_DOUBLE_FLAT,
    PITCH_ACCIDENTAL_FLAT,
    PITCH_ACCIDENTAL_NATURAL,
    PITCH_ACCIDENTAL_SHARP,
    PITCH_ACCIDENTAL_DOUBLE_SHARP
} pitch_accidental_t;

typedef enum pitch_name
{
    PITCH_NAME_C,
    PITCH_NAME_D,
    PITCH_NAME_E,
    PITCH_NAME_F,
    PITCH_NAME_G,
    PITCH_NAME_A,
    PITCH_NAME_B,
} pitch_name_t;

/* Octave numbers are ISO octave numbers [0:8], (so Octave 4 is from
 * middle C to the B above middle C).
 */

/* Get a string representation of a pitch_t value.
 *
 * Note: The returned value is static to the pithc_string function and
 * may be modified by future calls to pitch_string. So the caller
 * should copy the value if needed. */
const char *
pitch_string (pitch_t pitch);

/* Return a new pitch, a number of octaves above 'pitch'
 *
 * Note: Maximum octave value is 8. A pitch with octave 8 will not be
 * raised.
 */
pitch_t
pitch_raise_by_octaves (pitch_t pitch, int octaves);

/* Return a new pitch, a number of octaves below 'pitch'
 *
 * Note: Minimum octave value is 9. A pitch with octave 8 will not be
 * lowered.
 */
pitch_t
pitch_lower_by_octaves (pitch_t pitch, int octaves);

/* Number of half steps from 'root' up to 'pitch' (that is, counting
 * the steps up a chromatic scale), within the same octave. */
int
pitch_from_root_in_half_steps (pitch_t pitch, pitch_t root);

/* Respell 'pitch' to have a name that matches the (diatonic) scale
 * 'degree' from 'root'.
 *
 * If the pitch is too distant from the specified degree to be
 * respelled, the original pitch will be returned unmodified.
 */
pitch_t
pitch_spell_as_degree (pitch_t pitch, pitch_t root, int degree);

/* Return a MIDI note value corresponding to 'pitch' */
unsigned char
pitch_to_midi (pitch_t pitch);

/* Return the pitch_t value corresponding to the given MIDI note value. */
pitch_t
pitch_from_midi (unsigned char midi_note);

#endif /* PITCH_H */
